﻿/*
The MIT License

Copyright (c) 2009-2010. Carlos Guzmán Álvarez. http://chronoswpf.codeplex.com/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Windows;
using Chronos.Presentation.MVVM;
using Chronos.Presentation.MVVM.Services;

namespace Chronos.Presentation.Core.VirtualDesktops
{
    /// <summary>
    /// Interface for virtual desktop manager implementations
    /// </summary>
    public interface IVirtualDesktopManager
    {
        #region · Properties ·

        /// <summary>
        /// Gets the list of available virtual desktops.
        /// </summary>
        /// <value>The list of available virtual desktops.</value>
        ReadOnlyObservableCollection<IVirtualDesktop> VirtualDesktops
        {
            get;
        }

        /// <summary>
        /// Gets the active desktop windows.
        /// </summary>
        /// <value>The active desktop windows.</value>
        ICollection<ITrackableWindow> ActiveDesktopWindows
        {
            get;
        }

        /// <summary>
        /// Gets a value indicating whether there is an active virtual desktop.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if there are an active virtual desktop; otherwise, <c>false</c>.
        /// </value>
        bool HasDesktopActive
        {
            get;
        }

        #endregion

        #region · Methods ·

        /// <summary>
        /// Activates the default desktop
        /// </summary>
        void ActivateDefaultDesktop();

        /// <summary>
        /// Switches the active desktop
        /// </summary>
        void SwitchDesktop();

        /// <summary>
        /// Saves all the desktops to disk
        /// </summary>
        void SaveAllDesktops();

        /// <summary>
        /// Saves the active desktop to disk.
        /// </summary>
        void SaveCurrentDesktop();

        /// <summary>
        /// Creates a new shortcut with the given display name and target
        /// </summary>
        /// <param name="displayName"></param>
        /// <param name="target"></param>
        void CreateShortcut(string displayName, string target);

        /// <summary>
        /// Shows the specified shortcut.
        /// </summary>
        /// <param name="shortcut">The shortcut.</param>
        void Show(ShortcutViewModelBase shortcut);

        /// <summary>
        /// Shows the specified shortcut.
        /// </summary>
        /// <param name="shortcut">The shortcut.</param>
        /// <param name="position">The position.</param>
        void Show(ShortcutViewModelBase shortcut, Point position);

        /// <summary>
        /// Shows the specified widget.
        /// </summary>
        /// <param name="widget">The widget.</param>
        void Show(WidgetViewModel widget);

        /// <summary>
        /// Shows the specified widget.
        /// </summary>
        /// <param name="widget">The widget.</param>
        /// <param name="position">The position.</param>
        void Show(WidgetViewModel widget, Point position);

        /// <summary>
        /// Close the given element
        /// </summary>
        /// <param name="element"></param>
        void Close(ClosableViewModel element);

        /// <summary>
        /// Closes the element with the given Id
        /// </summary>
        /// <param name="id">The id.</param>
        void Close(Guid id);

        /// <summary>
        /// Closes all the elements
        /// </summary>
        /// <param name="id">The id.</param>
        void CloseAll();

        /// <summary>
        /// Shows the given <see cref="UIElement"/> as a normal Window
        /// </summary>
        /// <param name="window"></param>
        /// <remarks>The UIElement should inherit from <see cref="WindowElement"/></remarks>
        void Show(UIElement element);

        /// <summary>
        /// Shows the given <see cref="UIElement"/> as a modal Window
        /// </summary>
        /// <param name="window"></param>
        /// <remarks>The UIElement should inherit from <see cref="WindowElement"/></remarks>
        DialogResult ShowDialog(UIElement element);

        /// <summary>
        /// Closes the current dialog
        /// </summary>
        void CloseDialog();

        /// <summary>
        /// Restores the window with the given Id
        /// </summary>
        /// <param name="id"></param>
        void Restore(Guid id);

        /// <summary>
        /// Attach the given dependency object as a Desktop to the Window manager
        /// </summary>
        /// <param name="desktop"></param>
        void RegisterDesktop(DependencyObject d);

        /// <summary>
        /// Attach the given dependency object as a Modal Container to the Window manager
        /// </summary>
        /// <param name="desktop"></param>
        void RegisterModalContainer(DependencyObject d);

        #endregion
    }
}
