﻿/*
The MIT License

Copyright (c) 2009-2010. Carlos Guzmán Álvarez. http://chronoswpf.codeplex.com/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

using System;
using System.Diagnostics;
using System.IO;
using System.Threading.Tasks;
using Chronos.Presentation.MVVM.Messaging;
using Chronos.Presentation.MVVM.Services;
using nRoute.Components.Messaging;
using nRoute.ViewServices;

namespace Chronos.Presentation.MVVM
{
    /// <summary>
    /// External shortcut viewmodel class
    /// </summary>
    public sealed class ExternalShortcutViewModel : ShortcutViewModelBase
    {
        #region · Consts ·

        static readonly string DefaultIconStyle = "ArrowRight01IconStyle";

        #endregion

        #region · Fields ·

        private Uri     target;
        private string  iconStyle;

        #endregion

        #region · Properties ·

        /// <summary>
        /// Gets or sets the shortcut icon style.
        /// </summary>
        /// <value>The icon style.</value>
        public override string IconStyle
        {
            get
            {
                if (this.iconStyle == null)
                {
                    return DefaultIconStyle;
                }

                return this.iconStyle;
            }
            set
            {
                if (this.iconStyle != value)
                {
                    this.iconStyle = value;
                    this.NotifyPropertyChanged(() => IconStyle);
                }
            }
        }

        /// <summary>
        /// Gets or sets the shortcut target.
        /// </summary>
        /// <value>The shortcut target.</value>
        public Uri Target
        {
            get { return this.target; }
            set
            {
                if (this.target != value)
                {
                    this.target = value;
                    this.NotifyPropertyChanged(() => Target);
                }
            }
        }

        #endregion

        #region · Constructors ·

        /// <summary>
        /// Initializes a new instance of the <see cref="ExternalShortcutViewModel"/> class.
        /// </summary>
        public ExternalShortcutViewModel()
            : base()
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ExternalShortcutViewModel"/> class.
        /// </summary>
        public ExternalShortcutViewModel(string displayName)
            : base(displayName)
        {
        }

        #endregion

        #region · Command Execution ·

        /// <summary>
        /// Called when the <see cref="OpenCommand"/>  is executed.
        /// </summary>
        protected override void OnOpen()
        {
            Task task = Task.Factory.StartNew(
                () =>
                {
                    using (Process process = new Process())
                    {
                        process.StartInfo.FileName          = this.Target.ToString();
                        process.StartInfo.UseShellExecute   = true;
                        process.StartInfo.LoadUserProfile   = true;
                        process.Start();
                    }
                })
            .ContinueWith((t) =>
                {
                    Debug.WriteLine("I have observed a {0}", t.Exception.InnerException.GetType().Name);

                    IShowMessageViewService showMessageService = ViewServiceLocator.GetViewService<IShowMessageViewService>();

                    showMessageService.ButtonSetup  = DialogButton.Ok;
                    showMessageService.Caption      = "Problem with Shortcut";
                    showMessageService.Text         =
                        String.Format(
                            "An error has occurred while runninng the shortcut '{0}'",
                                this.DisplayName);

                    if (showMessageService.ShowMessage() == DialogResult.Yes)
                    {
                        Channel<DesktopElementActionInfo>.Publish(
                            new DesktopElementActionInfo { Id = this.Id, Action = DesktopElementAction.Close }, true);
                    }
                }, TaskContinuationOptions.OnlyOnFaulted);
        }

        #endregion
    }
}
