﻿/*
The MIT License

Copyright (c) 2009-2010. Carlos Guzmán Álvarez. http://chronoswpf.codeplex.com/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq.Expressions;
using System.Reflection;
using System.Windows.Threading;
using Chronos.ExtensionMethods;
using System.Diagnostics;

namespace Chronos.Presentation.MVVM
{
    /// <summary>
    /// Base class for observable objects viewmodels
    /// </summary>
    public abstract class ObservableObject : DispatcherObject, INotifyPropertyChanged
    {
        #region · Static Metadata Members ·

        /// <value>Cache of derived types property meta data. This meta data is used for notify property changes</value>
        static Dictionary<Type, List<PropertyInfo>> PropertyMetadata = new Dictionary<Type, List<PropertyInfo>>();

        #endregion

        #region · INotifyPropertyChanged Members ·

        public event PropertyChangedEventHandler PropertyChanged;

        #endregion

        #region · Constructors ·

        /// <summary>
        /// Initializes a new instance of the <see cref="ObservableObject"/> class.
        /// </summary>
        protected ObservableObject()
        {
        }

        #endregion

        #region · Protected Methods ·

        /// <summary>
        /// Initializes type metadata
        /// </summary>
        /// <param name="baseType"></param>
        protected void InitializeMetadata(Type baseType)
        {
            Debug.Assert(baseType != null);

            Type thisType = this.GetType();

            if (!PropertyMetadata.ContainsKey(thisType))
            {
                List<PropertyInfo> excludeList = new List<PropertyInfo>(
                    baseType.GetProperties(BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance));

                Type deriveType = this.GetType();

                List<PropertyInfo> includeList = new List<PropertyInfo>(
                    deriveType.GetProperties(BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance));

                includeList.RemoveAll(delegate(PropertyInfo pi)
                {
                    // Exclude base type and readonly properties
                    return excludeList.Contains(pi) || !pi.CanWrite;
                });

                PropertyMetadata.Add(thisType, includeList);
            }
        }

        /// <summary>
        /// Notifies all properties changed.
        /// </summary>
        protected void NotifyAllPropertiesChanged()
        {
            PropertyMetadata[this.GetType()].ForEach(delegate(PropertyInfo pi)
            {
                this.NotifyPropertyChanged(pi.Name);
            });
        }

        /// <summary>
        /// Notifies the property changed.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="property">The property.</param>
        protected void NotifyPropertyChanged<T>(Expression<Func<T>> property)
        {
            if (this.PropertyChanged != null)
            {
                this.PropertyChanged(this, property.CreateChangeEventArgs());
            }
        }

        /// <summary>
        /// Notifies the property changed.
        /// </summary>
        /// <param name="propertyName">Name of the property.</param>
        protected void NotifyPropertyChanged(string propertyName)
        {
            if (this.PropertyChanged != null)
            {
                this.PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }

        #endregion
    }
}
