﻿/*
The MIT License

Copyright (c) 2009-2010. Carlos Guzmán Álvarez. http://chronoswpf.codeplex.com/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

using System;
using System.Windows.Media.Imaging;
using Chronos.Presentation.MVVM.Messaging;
using Chronos.Presentation.MVVM.Services;
using nRoute.Components;
using nRoute.Components.Messaging;
using nRoute.Services;

namespace Chronos.Presentation.MVVM
{
    /// <summary>
    /// Base viewmodel class for workspace windows
    /// </summary>
    /// <typeparam name="TDataModel">The type of the data model.</typeparam>
    public class WorkspaceWindowViewModel<TDataModel> : WindowViewModel, ITrackableWindow
        where TDataModel: class, new()
    {
        #region · Events ·

        /// <summary>
        /// Occurs when view mode is changed.
        /// </summary>
        public event EventHandler ViewModeChanged;

        #endregion

        #region · Fields ·

        private ViewMode        viewMode;
        private TDataModel      dataModel;
        private BitmapSource    thumbnail;

        #region · Commands ·

        private ActionCommand restoreCommand;
        private ActionCommand addNewCommand;
        private ActionCommand editCommand;
        private ActionCommand deleteCommand;
        private ActionCommand saveCommand;
        private ActionCommand discardCommand;
        private ActionCommand inquiryCommand;
        private ActionCommand newWindowCommand;
        private ActionCommand createShortcutCommand;

        #endregion

        #endregion

        #region · Commands ·

        /// <summary>
        /// Gets the Add new command
        public ActionCommand AddNewCommand
        {
            get
            {
                if (this.addNewCommand == null)
                {
                    this.addNewCommand = new ActionCommand
                    (
                        () => OnAddNew(), 
                        () => CanAddNew()
                    );
                }

                return this.addNewCommand;
            }
        }

        /// <summary>
        /// Gets the Edit command
        /// </summary>
        public ActionCommand EditCommand
        {
            get
            {
                if (this.editCommand == null)
                {
                    this.editCommand = new ActionCommand
                    (
                        () => OnEdit(),
                        () => CanEdit()
                    );
                }

                return this.editCommand;
            }
        }

        /// <summary>
        /// Gets the Delete command
        /// </summary>
        public ActionCommand DeleteCommand
        {
            get
            {
                if (this.deleteCommand == null)
                {
                    this.deleteCommand = new ActionCommand
                    (
                        () => OnDelete(), 
                        () => CanDelete()
                    );
                }

                return this.deleteCommand;
            }
        }

        /// <summary>
        /// Gets the Save command
        /// </summary>
        public ActionCommand SaveCommand
        {
            get
            {
                if (this.saveCommand == null)
                {
                    this.saveCommand = new ActionCommand
                    (
                        () => OnSave(),
                        () => CanSave()
                    );
                }

                return this.saveCommand;
            }
        }

        /// <summary>
        /// Gets the Discard command
        /// </summary>
        public ActionCommand DiscardCommand
        {
            get
            {
                if (this.discardCommand == null)
                {
                    this.discardCommand = new ActionCommand
                    (
                        () => OnDiscard(),
                        () => CanDiscard()
                    );
                }

                return this.discardCommand;
            }
        }

        /// <summary>
        /// Gets the Refresh command
        /// </summary>
        public ActionCommand InquiryCommand
        {
            get
            {
                if (this.inquiryCommand == null)
                {
                    this.inquiryCommand = new ActionCommand
                    (
                        () => OnInquiryData(),
                        () => CanInquiryData()
                    );
                }

                return this.inquiryCommand;
            }
        }

        /// <summary>
        /// Gets the open new window command
        /// </summary>
        public ActionCommand NewWindowCommand
        {
            get
            {
                if (this.newWindowCommand == null)
                {
                    this.newWindowCommand = new ActionCommand
                    (
                        () => OnOpenNewWindow(),
                        () => CanOpenNewWindow()
                    );
                }

                return this.newWindowCommand;
            }
        }

        /// <summary>
        /// Gets the create shortcut command
        /// </summary>
        public ActionCommand CreateShortcutCommand
        {
            get
            {
                if (this.createShortcutCommand == null)
                {
                    this.createShortcutCommand = new ActionCommand
                    (
                        () => OnCreateShortcut(),
                        () => CanCreateShortcut()
                    );
                }

                return this.createShortcutCommand;
            }
        }

        #endregion

        #region · Properties ·

        /// <summary>
        /// Gets or sets the state of the smart part.
        /// </summary>
        /// <value>The state of the smart part.</value>
        public ViewMode ViewMode
        {
            get { return this.viewMode; }
            set
            {
                if (this.viewMode != value)
                {
                    this.viewMode = value;
                    this.OnViewModeChanged();
                }
            }
        }

        /// <summary>
        /// Gets the navigation URL
        /// </summary>
        public virtual string NavigationUrl
        {
            get { return String.Empty; }
        }

        #endregion

        #region · ITrackableWindow Members ·

        /// <summary>
        /// Gets or sets the window thumbnail.
        /// </summary>
        /// <value>The thumbnail.</value>
        public BitmapSource Thumbnail
        {
            get { return this.thumbnail; }
            set
            {
                if (this.thumbnail != value)
                {
                    this.thumbnail = value;
                    this.NotifyPropertyChanged(() => Thumbnail);
                }
            }
        }

        /// <summary>
        /// Gets the restore command.
        /// </summary>
        /// <value>The restore command.</value>
        public ActionCommand RestoreCommand
        {
            get
            {
                if (this.restoreCommand == null)
                {
                    this.restoreCommand = new ActionCommand(() => OnRestore());
                }

                return this.restoreCommand;
            }
        }

        #endregion

        #region · Protected Properties ·

        /// <summary>
        /// Gets or sets the data model.
        /// </summary>
        /// <value>The data model.</value>
        protected TDataModel DataModel
        {
            get { return this.dataModel; }
            set { this.dataModel = value; }
        }

        #endregion

        #region · Constructors ·

        /// <summary>
        /// Initializes a new instance of the <see cref="ViewModel&lt;TDataModel&gt;"/> class.
        /// </summary>
        protected WorkspaceWindowViewModel()
            : this(String.Empty)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ViewModel&lt;TDataModel&gt;"/> class.
        /// </summary>
        protected WorkspaceWindowViewModel(string displayName)
            : base(displayName)
        {
            this.InitializeMetadata(typeof(WorkspaceWindowViewModel<TDataModel>));

            this.DataModel  = new TDataModel();
            this.ViewMode   = ViewMode.ViewOnlyMode;
        }

        #endregion

        #region · Command Actions ·

        #region · Add New ·

        protected virtual bool CanAddNew()
        {
            return false;
        }

        protected virtual void OnAddNew()
        {
        }

        #endregion

        #region · Edit ·

        protected virtual bool CanEdit()
        {
            return false;
        }

        protected virtual void OnEdit()
        {
        }

        #endregion

        #region · Delete ·

        protected virtual bool CanDelete()
        {
            return false;
        }

        protected virtual void OnDelete()
        {
        }

        #endregion

        #region · Save ·

        protected virtual bool CanSave()
        {
            return false;
        }

        protected virtual void OnSave()
        {
        }

        #endregion

        #region · Discard ·

        protected virtual bool CanDiscard()
        {
            return false;
        }

        protected virtual void OnDiscard()
        {
        }

        #endregion

        #region · Inquiry ·

        protected virtual bool CanInquiryData()
        {
            return false;
        }

        protected virtual void OnInquiryData()
        {
        }

        #endregion

        #region · Open New Window Command ·

        protected virtual bool CanOpenNewWindow()
        {
            return !String.IsNullOrEmpty(this.NavigationUrl);
        }

        protected virtual void OnOpenNewWindow()
        {
            ServiceLocator.GetService<INavigationService>().NavigateTo(this.NavigationUrl);
        }

        #endregion

        #region · Create Shortcut Command ·

        protected virtual bool CanCreateShortcut()
        {
            return !String.IsNullOrEmpty(this.NavigationUrl);
        }

        protected virtual void OnCreateShortcut()
        {
#warning TODO: Publish a notification to create a new shortcut
            // ServiceLocator.GetService<IElementManager>().CreateShortcut(this.DisplayName, this.NavigationUrl);
        }

        #endregion

        #region · Close Action ·

        protected override bool CanCloseView()
        {
            return (this.ViewMode == ViewMode.ViewOnlyMode);
        }

        protected override void OnCloseView()
        {
            if (this.thumbnail != null)
            {
                this.thumbnail = null;
            }
            if (this.dataModel != null)
            {
                this.dataModel = null;
            }

            base.OnCloseView();
        }

        #endregion

        #region · Restore Action ·

        protected virtual void OnRestore()
        {
            Channel<DesktopElementActionInfo>.Publish(
                new DesktopElementActionInfo { Id = this.Id, Action = DesktopElementAction.Restore }, true);
        }

        #endregion

        #endregion

        #region · Protected Methods ·

        protected virtual void OnViewModeChanged()
        {
            this.UpdateAllowedUserActions();

            if (this.ViewModeChanged != null)
            {
                this.ViewModeChanged(this, new EventArgs());
            }

            this.NotifyAllPropertiesChanged();
        }

        protected virtual void UpdateAllowedUserActions()
        {
            this.AddNewCommand.RequeryCanExecute();
            this.EditCommand.RequeryCanExecute();
            this.DeleteCommand.RequeryCanExecute();
            this.SaveCommand.RequeryCanExecute();
            this.DiscardCommand.RequeryCanExecute();
            this.InquiryCommand.RequeryCanExecute();
            this.NewWindowCommand.RequeryCanExecute();
            this.CreateShortcutCommand.RequeryCanExecute();
        }

        #endregion
    }
}
