﻿/*
The MIT License

Copyright (c) 2009-2010. Carlos Guzmán Álvarez. http://chronoswpf.codeplex.com/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Windows.Input;
using Chronos.ExtensionMethods.Linq;
using Chronos.Presentation.Core.VirtualDesktops;
using Chronos.Presentation.MVVM;
using Chronos.Presentation.MVVM.Services;
using Chronos.Presentation.Windows;
using nRoute.Components;
using nRoute.Services;
using nRoute.SiteMaps;

namespace Chronos.Presentation.Widgets
{
    /// <summary>
    /// Navigator Widget ViewModel
    /// </summary>
    /// <remarks>
    /// http://www.codeproject.com/KB/WPF/TreeViewWithViewModel.aspx
    /// </remarks>
    public sealed class NavigatorWidgetViewModel : WidgetViewModel
    {
        #region · Fields ·

        private readonly ReadOnlyCollection<NavigationNode> firstGeneration;
        private readonly SiteMapNode                        rootOption;
        private List<NavigationNode>                        matchingOptions;
        private string                                      filterText;
        private ICommand                                    createWidgetCommand;
        private ICommand                                    navigateToCommand;

        #endregion

        #region · Commands ·

        /// <summary>
        /// Gets the create widget command.
        /// </summary>
        /// <value>The create widget command.</value>
        public ICommand CreateWidgetCommand
        {
            get
            {
                if (this.createWidgetCommand == null)
                {
                    this.createWidgetCommand = new ActionCommand(() => OnCreateWidget());
                }

                return this.createWidgetCommand;
            }
        }

        /// <summary>
        /// Gets the navigation command
        /// </summary>
        public ICommand NavigateToCommand
        {
            get
            {
                if (this.navigateToCommand == null)
                {
                    this.navigateToCommand = new ActionCommand
                    (
                        () => OnNavigateTo(),
                        () => CanNavigate()
                    );
                }

                return this.navigateToCommand;
            }
        }

        #endregion

        #region · Properties ·

        /// <summary>
        /// Gets the widget display name.
        /// </summary>
        /// <value>The name.</value>
        public override string DisplayName
        {
            get { return "Navigator"; }
            set { }
        }

        /// <summary>
        /// Gets the widget description.
        /// </summary>
        /// <value>The description.</value>
        public override string Description
        {
            get { return "Navigate and search application options"; }
            set { }
        }

        #region · FirstGeneration ·

        /// <summary>
        /// Returns a read-only collection containing the first application option
        /// in the tree, to which the TreeView can bind.
        /// </summary>
        public ReadOnlyCollection<NavigationNode> FirstGeneration
        {
            get
            {
                if (String.IsNullOrEmpty(this.filterText))
                {
                    return this.firstGeneration;
                }
                else
                {
                    return new ReadOnlyCollection<NavigationNode>(this.matchingOptions);
                }
            }
        }

        #endregion

        #region · FilterText ·

        /// <summary>
        /// Gets/sets a fragment of the name to filter for.
        /// </summary>
        public string FilterText
        {
            get { return this.filterText; }
            set
            {
                if (value != filterText)
                {
                    this.filterText = value;
                    this.PerformFilter();
                }
            }
        }

        #endregion

        #endregion

        #region · Constructor ·

        /// <summary>
        /// Initializes a new instance of the <see cref="NavigatorWidgetViewModel"/> class.
        /// </summary>
        public NavigatorWidgetViewModel()
            : base()
        {
            if (!DesignMode.IsInDesignMode)
            {
                this.rootOption      = SiteMapService.SiteMap.RootNode;
                this.matchingOptions = new List<NavigationNode>();
                this.firstGeneration = new ReadOnlyCollection<NavigationNode>(this.rootOption.ChildNodes.OfType<NavigationNode>().ToArray());
            }
        }

        #endregion

        #region · Command Actions ·

        /// <summary>
        /// Handles the create widget command
        /// </summary>
        private void OnCreateWidget()
        {
            ServiceLocator.GetService<IVirtualDesktopManager>().Show(new NavigatorWidgetViewModel());
        }

        /// <summary>
        /// Returns a value indcating whether the navigation command can be executed
        /// </summary>
        /// <returns></returns>
        private bool CanNavigate()
        {
            return this.matchingOptions != null && this.matchingOptions.Count == 1;
        }

        /// <summary>
        /// Handles the navigation command
        /// </summary>
        private void OnNavigateTo()
        {
            ServiceLocator.GetService<INavigationService>().NavigateTo(this.matchingOptions[0].Url); ;
        }

        #endregion

        #region · Search Logic ·

        private void PerformFilter()
        {
            this.VerifyMatchingOptionEnumerator();
        }

        private void VerifyMatchingOptionEnumerator()
        {
            // Clear current matching options
            this.matchingOptions.Clear();

            // Perform filter if needed
            if (!String.IsNullOrEmpty(this.filterText))
            {
                this.matchingOptions.AddRange(this.FindMatches(this.filterText));
            }

            // Notify changes
            this.NotifyPropertyChanged(() => FirstGeneration);
        }

        private List<NavigationNode> FindMatches(string filterText)
        {
            var nodes = from node in this.rootOption.ChildNodes.OfType<NavigationNode>().Traverse<NavigationNode>(node => node.ChildNodes.OfType<NavigationNode>())
                        where node.Title.StartsWith(filterText, StringComparison.OrdinalIgnoreCase) && (node.ChildNodes == null || node.ChildNodes.Count == 0)
                        select node;

            return nodes.ToList();
        }

        #endregion
    }
}
