﻿/*
The MIT License

Copyright (c) 2009-2010. Carlos Guzmán Álvarez. http://chronoswpf.codeplex.com/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

using System;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using Chronos.ExtensionMethods;
using Chronos.Presentation.Core.Windows;

namespace Chronos.Presentation.Windows.Controls
{
    /// <summary>
    /// Virtual desktop representation using a <see cref="Canvas"/> control
    /// </summary>
    public class Desktop : Canvas
    {
        #region · Dependency Properties ·

        /// <summary>
        /// Identifies the Id dependency property.
        /// </summary>
        public static readonly DependencyProperty IdProperty =
            DependencyProperty.Register("Id", typeof(Guid), typeof(Desktop),
                new FrameworkPropertyMetadata(Guid.NewGuid()));

        #endregion

        #region · Static Constructor ·

        /// <summary>
        /// Initializes the <see cref="Desktop"/> class.
        /// </summary>
        static Desktop()
        {
            FocusManager.IsFocusScopeProperty.OverrideMetadata(
                typeof(Desktop), new FrameworkPropertyMetadata(true));

            KeyboardNavigation.ControlTabNavigationProperty.OverrideMetadata(
                typeof(Desktop), new FrameworkPropertyMetadata(KeyboardNavigationMode.Cycle));
        }

        #endregion

        #region · Properties ·

        /// <summary>
        /// Gets or sets the element id.
        /// </summary>
        /// <value>The id.</value>
        public Guid Id
        {
            get { return (Guid)base.GetValue(IdProperty); }
            set { base.SetValue(IdProperty, value); }
        }

        #endregion

        #region · Constructor ·

        /// <summary>
        /// Initializes a new instance of the <see cref="Desktop"/> class.
        /// </summary>
        public Desktop()
            : base()
        {
        }

        #endregion

        #region · Methods ·

        /// <summary>
        /// Adds the new desktop element.
        /// </summary>
        /// <param name="newElement">The new desktop element.</param>
        /// <param name="position">The position.</param>
        public void AddElement(DesktopElement newElement, Point position)
        {
            newElement.Move(Math.Max(0, position.X), Math.Max(0, position.Y));
            newElement.SetZIndex(0);

            this.AddElement(newElement);
        }

        /// <summary>
        /// Adds the new desktop element.
        /// </summary>
        /// <param name="newElement">The new desktop element.</param>
        public void AddElement(DesktopElement newElement)
        {
            this.Children.Add(newElement);
        }

        /// <summary>
        /// Removes the given element from the desktop.
        /// </summary>
        /// <param name="element">The element.</param>
        public void RemoveElement(DesktopElement element)
        {
            this.Children.Remove(element);
        }

        #endregion

        #region  · Mouse Handling Methods ·

        /// <summary>
        /// Invoked when an unhandled <see cref="E:System.Windows.Input.Mouse.MouseDown"/> attached event 
        /// reaches an element in its route that is derived from this class. 
        /// Implement this method to add class handling for this event.
        /// </summary>
        /// <param name="e">The <see cref="T:System.Windows.Input.MouseButtonEventArgs"/> 
        /// that contains the event data. This event data reports details about the mouse button that was pressed and the handled state.
        /// </param>
        protected override void OnMouseDown(MouseButtonEventArgs e)
        {            
            if (Object.ReferenceEquals(e.Source,  this))
            {
                // Deactivate all Desktop elements
                this.DeactivateAll();

                FocusHelper.Focus(this);

                e.Handled = true;
            }

            base.OnMouseDown(e);
        }

        #endregion

        #region · Internal Methods ·

        /// <summary>
        /// Called when a <see cref="DesktopElement"/> gets activated.
        /// </summary>
        /// <param name="id">The id.</param>
        internal void OnActivatedElement(Guid id)
        {
            this.Children.OfType<DesktopElement>().Where(e => e.Id != id).ToList().ForEach(element => element.Deactivate());
        }

        #endregion

        #region · Private Methods ·

        private void DeactivateAll()
        {
            this.Children.OfType<IActiveAware>().ToList().ForEach(element => element.Deactivate());
        }

        #endregion
    }
}
