﻿/*
The MIT License

Copyright (c) 2009-2010. Carlos Guzmán Álvarez. http://chronoswpf.codeplex.com/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Windows;
using System.Windows.Threading;
using Chronos.ExtensionMethods;
using Chronos.Presentation.Core.VirtualDesktops;
using Chronos.Presentation.MVVM;
using Chronos.Presentation.Windows.Controls;
using nRoute.Components.Messaging;

namespace Chronos.Presentation.Windows
{
    /// <summary>
    /// Virtual desktop class
    /// </summary>
    public sealed class VirtualDesktop : DispatcherObject, IVirtualDesktop
    {
        #region · Static Methods ·

        private T CreateDesktopElement<T>(ClosableViewModel vm) where T: DesktopElement, new()
        {
            return new T { DataContext = vm, Content = vm };
        }

        #endregion

        #region · Fields ·

        private Desktop                                         desktop;
        private ReadOnlyObservableCollection<ITrackableWindow>  activeWindowsWrapper;
        private ObservableCollection<ITrackableWindow>          activeWindows;
        private bool                                            hasBeenActivated;

        #endregion

        #region · Properties ·

        public Guid Id
        {
            get
            {
                if (this.desktop != null)
                {
                    return this.desktop.Id;
                }

                return Guid.Empty;
            }
        }

        /// <summary>
        /// Gets the list of active Windows
        /// </summary>
        public ReadOnlyObservableCollection<ITrackableWindow> ActiveWindows
        {
            get
            {
                if (this.activeWindowsWrapper == null)
                {
                    this.activeWindowsWrapper = new ReadOnlyObservableCollection<ITrackableWindow>(this.Windows);
                }

                return this.activeWindowsWrapper;
            }
        }

        #endregion

        #region · Private Properties ·

        private ObservableCollection<ITrackableWindow> Windows
        {
            get
            {
                if (this.activeWindows == null)
                {
                    this.activeWindows = new ObservableCollection<ITrackableWindow>();
                }

                return this.activeWindows;
            }
        }

        #endregion

        #region · Constructors ·

        /// <summary>
        /// Initializes a new instance of <see cref="VirtualDesktop"/> class
        /// </summary>
        internal VirtualDesktop(Desktop desktop)
        {
            this.desktop = desktop;
        }

        #endregion

        #region · Desktop Actions ·

        /// <summary>
        /// Activates the desktop instance
        /// </summary>
        public void Activate()
        {
            this.InvokeAsynchronouslyInBackground(
                () =>
                {
                    if (!this.hasBeenActivated)
                    {
                        this.Load();
                        this.hasBeenActivated = true;
                    }

                    this.desktop.Visibility = Visibility.Visible;

                    Channel<ActiveDesktopChangedInfo>.Publish(new ActiveDesktopChangedInfo(this), true);
                });
        }

        /// <summary>
        /// Deactivates the desktop instance
        /// </summary>
        public void Deactivate()
        {
            this.Invoke(
                () =>
                {
                    this.desktop.Visibility = Visibility.Hidden;
                });
        }

        /// <summary>
        /// Saves desktop contents to disk
        /// </summary>
        public void Save()
        {
            this.Invoke(
                () =>
                {
                    DesktopSerializer.Save(this.desktop, this.GetXamlFilename());
                });
        }

        #endregion

        #region · Window Methods ·

        /// <summary>
        /// Shows the given window instance
        /// </summary>
        /// <param name="element">A <see cref="WindowElement"/> instance</param>
        public void Show(UIElement element)
        {
            Debug.Assert(element is WindowElement);

            WindowElement window = element as WindowElement;

            this.InvokeAsynchronouslyInBackground(
                () =>
                {
                    this.desktop.AddElement(window);
                    window.Parent = this.desktop;
                    window.Show();

                    // Add the window to the ActiveWindows collection
                    if (!window.IsModal &&
                        window.DataContext is ITrackableWindow)
                    {
                        this.InvokeAsynchronouslyInBackground(
                            delegate
                            {
                                ITrackableWindow vm = window.DataContext as ITrackableWindow;
                                vm.Thumbnail = window.CaptureScreen();

                                this.Windows.Add(vm);
                            });
                    }
                });
        }

        /// <summary>
        /// Restores the window with the given Id
        /// </summary>
        /// <param name="id"></param>
        public void Restore(Guid id)
        {
            this.InvokeAsynchronouslyInBackground(
                () =>
                {
                    WindowElement window = this.desktop.Children.OfType<WindowElement>().Where(wc => wc.Id == id).SingleOrDefault();

                    if (window != null)
                    {
                        window.WindowState = WindowState.Normal;
                        window.Activate();
                    }
                });
        }

        #endregion

        #region · Shortcut Methods ·

        /// <summary>
        /// Creates a new shortcut with the given display name and target
        /// </summary>
        /// <param name="displayName"></param>
        /// <param name="target"></param>
        public void CreateShortcut(string displayName, string target)
        {
            InternalShortcutViewModel vm = new InternalShortcutViewModel
            {
                DisplayName = displayName,
                Target      = target
            };
            
            this.Show(vm);
        }

        /// <summary>
        /// Shows the specified shortcut.
        /// </summary>
        /// <param name="shortcut">The shortcut.</param>
        public void Show(ShortcutViewModelBase shortcut)
        {
            this.Show(shortcut, new Point());
        }

        /// <summary>
        /// Shows the specified shortcut.
        /// </summary>
        /// <param name="shortcut">The shortcut.</param>
        /// <param name="position">The position.</param>
        public void Show(ShortcutViewModelBase shortcut, Point position)
        {
            this.InvokeAsynchronouslyInBackground(
                () =>
                {
                    this.desktop.AddElement(CreateDesktopElement<ShortcutElement>(shortcut), position);
                });
        }

        #endregion

        #region · Widget Methods ·

        /// <summary>
        /// Shows the specified widget.
        /// </summary>
        /// <param name="widget">The widget.</param>
        public void Show(WidgetViewModel widget)
        {
            this.Show(widget, new Point());
        }

        /// <summary>
        /// Shows the specified widget.
        /// </summary>
        /// <param name="widget">The widget.</param>
        /// <param name="position">The position.</param>
        public void Show(WidgetViewModel widget, Point position)
        {
            this.InvokeAsynchronously(
                () =>
                {
                    this.desktop.AddElement(CreateDesktopElement<WidgetElement>(widget), position);
                });
        }

        #endregion

        #region · Common Methods ·

        /// <summary>
        /// Closes the given element with the given Id
        /// </summary>
        /// <param name="id">The id.</param>
        public void Close(ClosableViewModel vm)
        {
            this.Close(vm.Id);
        }

        /// <summary>
        /// Closes the given element with the given Id
        /// </summary>
        /// <param name="id">The id.</param>
        public void Close(Guid id)
        {
            DesktopElement element = this.desktop.Children.OfType<DesktopElement>().Where(x => x.Id == id).FirstOrDefault();

            if (element != null)
            {
                this.Close(element);
            }
        }

        /// <summary>
        /// Closes the given element
        /// </summary>
        /// <param name="element">The desktop element</param>
        public void Close(DesktopElement element)
        {
            Debug.Assert(element != null, "element");

            this.Invoke(
                () =>
                {
                    if (element is WindowElement)
                    {
                        WindowElement window = element as WindowElement;

                        if (window.DataContext is ITrackableWindow)
                        {
                            this.Windows.Remove(this.Windows.Where(wnd => wnd.Id == window.Id).SingleOrDefault());
                        }
                    }

                    this.desktop.RemoveElement(element);

                    element.Close();
                    element = null;
                });
        }

        /// <summary>
        /// Closes all
        /// </summary>
        public void CloseAll()
        {
            this.Invoke(
                () =>
                {
                    List<DesktopElement> elements = this.desktop.Children.OfType<DesktopElement>().ToList();

                    if (elements != null)
                    {
                        foreach (DesktopElement element in elements)
                        {
                            this.Close(element);
                        }

                        elements.Clear();
                        elements = null;

                        this.hasBeenActivated = false;
                    }
                });
        }

        #endregion

        #region · Private Methods ·

        /// <summary>
        /// Loads desktop contents from disk
        /// </summary>
        private void Load()
        {
            this.InvokeAsynchronouslyInBackground(
                () =>
                {
                    DesktopSerializer.Load(this.desktop, this.GetXamlFilename());
                });
        }

        private string GetXamlFilename()
        {
            return String.Format("{0}.xaml", this.desktop.Name);
        }

        private void RemoveElement(DesktopElement element)
        {
            this.InvokeAsynchronouslyInBackground(
                () =>
                {
                    this.desktop.RemoveElement(element);
                });
        }

        #endregion
    }
}
