﻿/*
The MIT License

Copyright (c) 2009-2010. Carlos Guzmán Álvarez. http://chronoswpf.codeplex.com/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Threading;
using Chronos.ExtensionMethods;
using Chronos.Presentation.Core.VirtualDesktops;
using Chronos.Presentation.MVVM;
using Chronos.Presentation.MVVM.Services;
using Chronos.Presentation.Windows.Controls;
using nRoute.Components.Composition;
using nRoute.Services;

namespace Chronos.Presentation.Windows
{
    [MapService(typeof(IVirtualDesktopManager),
        InitializationMode=InitializationMode.OnDemand,
        Lifetime=InstanceLifetime.Singleton)]
    public sealed class VirtualDesktopManager : DispatcherObject, IVirtualDesktopManager
    {
        #region · Attached Properties ·

        /// <summary>
        /// Identifies the IsDesktopCanvas dependency property.
        /// </summary>
        public static readonly DependencyProperty IsDesktopProperty =
            DependencyProperty.RegisterAttached("IsDesktop", typeof(bool), typeof(VirtualDesktopManager),
                new FrameworkPropertyMetadata((bool)false,
                    new PropertyChangedCallback(OnIsDesktop)));

        /// <summary>
        /// Identifies the IsModalContainer dependency property.
        /// </summary>
        public static readonly DependencyProperty IsIsModalContainerProperty =
            DependencyProperty.RegisterAttached("IsModalContainer", typeof(bool), typeof(VirtualDesktopManager),
                new FrameworkPropertyMetadata((bool)false,
                    new PropertyChangedCallback(OnIsModalContainer)));

        #endregion

        #region · Attached Property Get/Set Methods ·

        /// <summary>
        /// Gets the value of the IsDesktopCanvas attached property
        /// </summary>
        /// <param name="d"></param>
        /// <returns></returns>
        public static bool GetIsDesktop(DependencyObject d)
        {
            return (bool)d.GetValue(IsDesktopProperty);
        }

        /// <summary>
        /// Sets the value of the IsDesktopCanvas attached property
        /// </summary>
        /// <param name="d"></param>
        /// <param name="value"></param>
        public static void SetIsDesktop(DependencyObject d, bool value)
        {
            d.SetValue(IsDesktopProperty, value);
        }

        /// <summary>
        /// Gets the value of the IsModalContainer attached property
        /// </summary>
        /// <param name="d"></param>
        /// <returns></returns>
        public static bool GetIsModalContainer(DependencyObject d)
        {
            return (bool)d.GetValue(IsIsModalContainerProperty);
        }

        /// <summary>
        /// Sets the value of the IsModalContainer attached property
        /// </summary>
        /// <param name="d"></param>
        /// <param name="value"></param>
        public static void SetIsModalContainer(DependencyObject d, bool value)
        {
            d.SetValue(IsIsModalContainerProperty, value);
        }

        #endregion

        #region · Attached Properties Callbacks ·

        private static void OnIsDesktop(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d != null)
            {
                IVirtualDesktopManager service;

                if (ServiceLocator.TryGetService<IVirtualDesktopManager>(out service))
                {
                    service.RegisterDesktop(d);
                }
            }
        }

        private static void OnIsModalContainer(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d != null)
            {
                IVirtualDesktopManager service;

                if (ServiceLocator.TryGetService<IVirtualDesktopManager>(out service))
                {
                    service.RegisterModalContainer(d);
                }
            }
        }

        #endregion

        #region · Fields ·

        private ObservableCollection<IVirtualDesktop>           virtualDesktops;
        private ReadOnlyObservableCollection<IVirtualDesktop>   virtualDesktopsWrapper;
        private IVirtualDesktop                                 activeDesktop;

        #endregion

        #region · Properties ·

        public ReadOnlyObservableCollection<IVirtualDesktop> VirtualDesktops
        {
            get
            {
                if (this.virtualDesktopsWrapper == null)
                {
                    this.virtualDesktopsWrapper = new ReadOnlyObservableCollection<IVirtualDesktop>(this.Desktops);
                }

                return this.virtualDesktopsWrapper;
            }
        }

        public ICollection<ITrackableWindow> ActiveDesktopWindows
        {
            get { return this.ActiveDesktop.ActiveWindows; }
        }

        public bool HasDesktopActive
        {
            get { return (this.ActiveDesktop != null); }
        }

        #endregion

        #region · Private Properties ·

        private ObservableCollection<IVirtualDesktop> Desktops
        {
            get
            {
                if (this.virtualDesktops == null)
                {
                    this.virtualDesktops = new ObservableCollection<IVirtualDesktop>();
                }

                return this.virtualDesktops;
            }
        }

        private IVirtualDesktop ActiveDesktop
        {
            get { return this.activeDesktop; }
            set
            {
                if (this.activeDesktop != value)
                {
                    this.activeDesktop = value;
                }
            }
        }

        #endregion

        #region · Constructors ·

        public VirtualDesktopManager()
        {
        }

        #endregion

        #region · Virtual Desktop Actions ·

        /// <summary>
        /// Activates the default desktop
        /// </summary>
        public void ActivateDefaultDesktop()
        {
            Debug.Assert(this.Desktops.Count > 0, "There are no desktop registered");

            this.ActiveDesktop = this.Desktops.First();
            this.ActiveDesktop.Activate();
        }

        /// <summary>
        /// Switches the active desktop
        /// </summary>
        public void SwitchDesktop()
        {
            Debug.Assert(this.Desktops.Count > 0, "There are no desktop registered");

            IVirtualDesktop currentDesktop = this.ActiveDesktop;

            this.Desktops.ToList().ForEach(vd => vd.Deactivate());
            this.ActiveDesktop = null;

            IVirtualDesktop desktop = this.Desktops.Where(vd => vd.Id != currentDesktop.Id).FirstOrDefault();

            if (desktop != null)
            {
                this.ActiveDesktop = desktop;
                this.ActiveDesktop.Activate();
            }
            else
            {
                this.ActivateDefaultDesktop();
            }
        }

        /// <summary>
        /// Saves the active desktop to disk
        /// </summary>
        public void SaveCurrentDesktop()
        {
            Debug.Assert(this.HasDesktopActive);

            this.ActiveDesktop.Save();
        }

        public void SaveAllDesktops()
        {
            this.Desktops.ToList().ForEach(vd => vd.Save());
        }

        #endregion

        #region · Window Methods ·

        /// <summary>
        /// Shows the given window instance
        /// </summary>
        /// <param name="element">A <see cref="WindowElement"/> instance</param>
        public void Show(UIElement element)
        {
            Debug.Assert(this.HasDesktopActive);

            this.ActiveDesktop.Show(element);
        }

        /// <summary>
        /// Shows the given window instance as a modal dialog
        /// </summary>
        /// <param name="element">A <see cref="WindowElement"/> instance</param>
        public DialogResult ShowDialog(UIElement element)
        {
            Debug.Assert(element is WindowElement);

            WindowElement   window          = element as WindowElement;
            DialogResult    dialogResult    = DialogResult.None;

            this.Invoke(
                delegate
                {
                    try
                    {
                        dialogResult = window.ShowDialog();
                    }
                    finally
                    {
                        if (window != null)
                        {
                            window.Close();
                            window = null;
                        }
                    }
                });

            return dialogResult;
        }

        /// <summary>
        /// Closes the current dialog
        /// </summary>
        public void CloseDialog()
        {
            this.Invoke(
                delegate
                {
                    WindowElement window = WindowElement.ModalContainerPanel.Children.OfType<WindowElement>().FirstOrDefault();

                    if (window != null)
                    {
                        window.Hide();
                    }
                });
        }

        /// <summary>
        /// Restores the window with the given Id
        /// </summary>
        /// <param name="id"></param>
        public void Restore(Guid id)
        {
            Debug.Assert(this.HasDesktopActive);

            this.ActiveDesktop.Restore(id);
        }

        #endregion

        #region · Shortcut Methods ·

        /// <summary>
        /// Creates a new shortcut with the given display name and target
        /// </summary>
        /// <param name="displayName"></param>
        /// <param name="target"></param>
        public void CreateShortcut(string displayName, string target)
        {
            Debug.Assert(this.HasDesktopActive);

            this.ActiveDesktop.CreateShortcut(displayName, target);
        }

        /// <summary>
        /// Shows the specified shortcut.
        /// </summary>
        /// <param name="shortcut">The shortcut.</param>
        public void Show(ShortcutViewModelBase shortcut)
        {
            Debug.Assert(this.HasDesktopActive);

            this.ActiveDesktop.Show(shortcut);
        }

        /// <summary>
        /// Shows the specified shortcut.
        /// </summary>
        /// <param name="shortcut">The shortcut.</param>
        /// <param name="position">The position.</param>
        public void Show(ShortcutViewModelBase shortcut, Point position)
        {
            Debug.Assert(this.HasDesktopActive);

            this.ActiveDesktop.Show(shortcut, position);
        }

        #endregion

        #region · Widget Methods ·

        /// <summary>
        /// Shows the specified widget.
        /// </summary>
        /// <param name="widget">The widget.</param>
        public void Show(WidgetViewModel widget)
        {
            Debug.Assert(this.HasDesktopActive);

            this.ActiveDesktop.Show(widget);
        }

        /// <summary>
        /// Shows the specified widget.
        /// </summary>
        /// <param name="widget">The widget.</param>
        /// <param name="position">The position.</param>
        public void Show(WidgetViewModel widget, Point position)
        {
            Debug.Assert(this.HasDesktopActive);

            this.ActiveDesktop.Show(widget, position);
        }

        #endregion

        #region · Common Methods ·

        /// <summary>
        /// Closes the given element with the given Id
        /// </summary>
        /// <param name="id">The id.</param>
        public void Close(ClosableViewModel vm)
        {
            Debug.Assert(this.HasDesktopActive);

            this.ActiveDesktop.Close(vm);
        }

        /// <summary>
        /// Closes the given element with the given Id
        /// </summary>
        /// <param name="id">The id.</param>
        public void Close(Guid id)
        {
            Debug.Assert(this.HasDesktopActive);

            this.ActiveDesktop.Close(id);
        }

        /// <summary>
        /// Closes all
        /// </summary>
        public void CloseAll()
        {
            Debug.Assert(this.HasDesktopActive);

            this.Desktops.ToList().ForEach(vd => vd.CloseAll());
            this.DeactivateAll();
        }

        #endregion

        #region · Desktop registration Methods ·

        /// <summary>
        /// Attach the given dependency object as a Desktop to the Window manager
        /// </summary>
        /// <param name="desktop"></param>
        public void RegisterDesktop(DependencyObject d)
        {
            Debug.Assert(d != null, "do");
            Debug.Assert(d is Desktop, "do is not an instance of Desktop");

            Desktop desktop = d as Desktop;

            if (desktop != null)
            {
                if (this.Desktops.Where(vd => vd.Id == desktop.Id).Count() == 0)
                {
                    this.Desktops.Add(new VirtualDesktop(desktop));
                }
            }
            else
            {
                Debug.WriteLine("d is not a valid Desktop instance");
            }
        }

        /// <summary>
        /// Attach the given dependency object as a Modal Container to the Window manager
        /// </summary>
        /// <param name="desktop"></param>
        public void RegisterModalContainer(DependencyObject d)
        {
            Canvas container = d as Canvas;

            if (container != null)
            {
                WindowElement.ModalContainerPanel = container;
            }
            else
            {
                Debug.WriteLine("d is not a valid Canvas instance");
            }
        }

        #endregion

        #region · Private Methods ·

        private void DeactivateAll()
        {
            this.Desktops.ToList().ForEach(vd => vd.Deactivate());
        }

        #endregion
    }
}
