/*
The MIT License

Copyright (c) 2009-2010. Carlos Guzmn lvarez. http://chronoswpf.codeplex.com/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

using System;
using System.ComponentModel;
using System.Linq.Expressions;

namespace Chronos.Presentation.ViewModel
{
    /// <summary>
    /// Represents a property state
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public sealed class PropertyState<T> 
        : ObservableObject where T: class
    {
        #region  PropertyChangedEventArgs Cached Instances 

        private static readonly PropertyChangedEventArgs IsEditableChangedArgs = CreateArgs<PropertyState<T>>(x => x.IsEditable);
        private static readonly PropertyChangedEventArgs IsReadOnlyChangedArgs = CreateArgs<PropertyState<T>>(x => x.IsReadOnly);

        #endregion

        #region  Fields 

        private string	name;
        private bool    isEditable;
        
        #endregion

        #region  Properties 

        /// <summary>
        /// Gets the property name
        /// </summary>
        public string PropertyName
        {
            get { return this.name; }
        }

        /// <summary>
        /// Gets or sets a value that indicates whether the property is editable
        /// </summary>
        public bool IsEditable
        {
            get { return this.isEditable; }
            set
            {
                this.isEditable = value;
                this.NotifyPropertyChanged(IsEditableChangedArgs);
                this.NotifyPropertyChanged(IsReadOnlyChangedArgs);
            }
        }

        /// <summary>
        /// Gets or sets a value that indicates whether the property is readonly
        /// </summary>
        public bool IsReadOnly
        {
            get { return !this.isEditable; }
        }

        #endregion

        #region  Constructors 

        /// <summary>
        /// Initializes a new instance of the <see cref="PropertyState&lt;T&gt;"/> class.
        /// </summary>
        /// <param name="propertyName">Name of the property.</param>
        public PropertyState(string propertyName)
        {
            this.name = propertyName;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PropertyState&lt;T&gt;"/> class.
        /// </summary>
        /// <param name="propertyExpression">The property expression.</param>
        public PropertyState(Expression<Func<T, object>> propertyExpression)
        {
            this.name = propertyExpression.GetPropertyName();
        }

        #endregion
    }
}