﻿/*
The MIT License

Copyright (c) 2009-2010. Carlos Guzmán Álvarez. http://chronoswpf.codeplex.com/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

using System;
using System.Windows;
using Chronos.Authentication;
using Chronos.Model;
using Chronos.Presentation.Core.VirtualDesktops;
using Chronos.Presentation.Core.Windows;
using Chronos.Presentation.ViewModel;
using nRoute.Components.Messaging;
using nRoute.Services;

namespace Chronos.ViewModel
{
    /// <summary>
    /// Login view view model class
    /// </summary>
    public sealed class LoginViewModel
        : WindowViewModel<UserLogin>
    {
        #region · Data Properties ·

        /// <summary>
        /// Gets or sets the user name
        /// </summary>
        public string UserId
        {
            get { return this.Entity.UserId; }
            set
            {
                if (this.Entity.UserId != value)
                {
                    this.Entity.UserId = value;
                    this.NotifyPropertyChanged(() => UserId);

                    this.InquiryCommand.RequeryCanExecute();
                }
            }
        }

        /// <summary>
        /// Gets or sets the password
        /// </summary>
        public string Password
        {
            get { return this.Entity.Password; }
            set
            {
                if (this.Entity.Password != value)
                {
                    this.Entity.Password = value;
                    this.NotifyPropertyChanged(() => Password);

                    this.InquiryCommand.RequeryCanExecute();
                }
            }
        }        

        #endregion

        #region · Constructors ·

        /// <summary>
        /// Initializes a new instance of the <see cref="LoginViewModel"/> class
        /// </summary>
        public LoginViewModel() 
            : base()
        {
        }

        #endregion

        #region · Overriden Methods ·

        public override bool CanClose()
        {
            return (this.ViewMode != ViewModeType.Busy);
        }

        public override void Close()
        {
            this.GetService<IVirtualDesktopManager>().CloseDialog();
            Application.Current.Shutdown();
        }

        protected override void InitializePropertyStates()
        {
            this.PropertyStates.Add(e => e.UserId);
            this.PropertyStates.Add(e => e.Password);
        }

        protected override void OnViewModeChanged()
        {
            base.OnViewModeChanged();

            if (this.PropertyStates.Count > 0)
            {
                this.PropertyStates[x => x.UserId].IsEditable   = (this.ViewMode != ViewModeType.Busy);
                this.PropertyStates[x => x.Password].IsEditable = (this.ViewMode != ViewModeType.Busy);
            }
        }

        #endregion

        #region · Command Actions ·

        protected override bool CanInquiryData()
        {
            return (!String.IsNullOrEmpty(this.UserId) &&
                    !String.IsNullOrEmpty(this.Password)    && 
                    this.ViewMode != ViewModeType.Busy);
        }

        protected override void OnInquiryAction(InquiryActionResult<UserLogin> result)
        {
            result.Data     = this.Entity;
            result.Result   = InquiryActionResultType.DataFetched;
        }

        protected override void OnInquiryActionComplete(InquiryActionResult<UserLogin> result)
        {
            if (result.Result == InquiryActionResultType.DataFetched)
            {
                Channel<AuthenticationInfo>.Public.OnNext(
                    new AuthenticationInfo
                    {
                        Action = AuthenticationAction.LoggedIn,
                        UserId = this.UserId
                    }, true);

                ServiceLocator.GetService<IVirtualDesktopManager>().CloseDialog();
            }
            else if (result.Result == InquiryActionResultType.DataNotFound)
            {
                this.NotificationMessage = "Username and password do not match.";

                this.ViewMode = ViewModeType.Default;
            }
        }

        #endregion
    }
}
